/**
 * Copyright (c) 2010 Alois Glomann gnalff@lucubrators.net
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package net.lucubrators.honeycomb.defaultimpl.reflect;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Default implementation of a {@link MethodResolver}.
 * Additional feature is method caching by controller class.
 * 
 * @author gnalFF
 * 
 */
public class DefaultMethodResolver implements MethodResolver {

	private ConcurrentHashMap<ResolverCacheItem, MethodHolder<?>> cache;

	/**
	 * Default Constructor
	 */
	public DefaultMethodResolver() {
		cache = new ConcurrentHashMap<ResolverCacheItem, MethodHolder<?>>();
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T extends Annotation> MethodHolder<T> resolveMethod(final Object from, final Class<T> annotationClass) {
		ResolverCacheItem cacheItem = new ResolverCacheItem(annotationClass, from);
		if (cache.containsKey(cacheItem)) {
			return (MethodHolder<T>) cache.get(cacheItem);
		}
		Class<? extends Object> controllerClass = from.getClass();
		Method[] methods = controllerClass.getMethods();
		for (Method method : methods) {
			Annotation[] annotations = method.getAnnotations();
			for (Annotation annotation : annotations) {
				if (annotation.annotationType().equals(annotationClass)) {
					MethodHolder<T> methodHolder = new MethodHolder<T>(from, method, (T) annotation);
					MethodHolder<T> cachedResult = (MethodHolder<T>) cache.putIfAbsent(cacheItem, methodHolder);
					if (cachedResult == null) {
						return methodHolder;
					}
					return cachedResult;
				}
			}
		}
		return null;
	}

	private class ResolverCacheItem {
		private final Object container;
		private final Class<?> annotationType;

		public ResolverCacheItem(Class<?> annotationType, Object container) {
			this.annotationType = annotationType;
			this.container = container;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + getOuterType().hashCode();
			result = prime * result + ((annotationType == null) ? 0 : annotationType.hashCode());
			result = prime * result + ((container == null) ? 0 : container.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			ResolverCacheItem other = (ResolverCacheItem) obj;
			if (!getOuterType().equals(other.getOuterType()))
				return false;
			if (annotationType == null) {
				if (other.annotationType != null)
					return false;
			} else if (!annotationType.equals(other.annotationType))
				return false;
			if (container == null) {
				if (other.container != null)
					return false;
			} else if (container != (other.container))
				return false;
			return true;
		}

		private DefaultMethodResolver getOuterType() {
			return DefaultMethodResolver.this;
		}

	}

}
